


import json
from typing import Any, Dict, List, Optional
from requests import Response
from tabulate import tabulate

from .utils import api_get
from .auth import require_auth


@require_auth
def attackcategories(access_token: str, json_format: bool = False) -> Response:
    res = api_get("https://api.sandbox.mindgard.ai/api/v1/attacks/categories", access_token)
    category_names: List[str] = list(map(lambda x: x["category"], res.json()))
    print(json.dumps(res.json(), indent=2)) if json_format else print("\n".join(category_names))
    return res


def display_attacks_results(data: List[Dict[str, Any]]) -> None:
    display_data: List[Dict[str, Any]] = []
    for d in data:
        row = {k: v for k, v in d.items() if k not in ["stacktrace", "submitted_at_unix", "run_at_unix"]}

        display_data.append(row)
    print(tabulate(display_data, headers="keys"))


@require_auth
def get_attacks(access_token: str, json_format: bool = False, attack_id: Optional[str] = None) -> Response:
    url = f"https://api.sandbox.mindgard.ai/api/v1/results/{attack_id}" if attack_id else "https://api.sandbox.mindgard.ai/api/v1/users/experiments"
    res = api_get(url, access_token)
    data: List[Dict[str, Any]] = res.json() if isinstance(res.json(), list) else [res.json()]
    if json_format:
        print(json.dumps(res.json()))
    else:
        print(json.dumps(res.json(), indent=2)) if attack_id else display_attacks_results(data)
    return res




VERSION: str = "0.15.0"
REPOSITORY_URL: str = f"https://pypi.org/pypi/mindgard/json"
AUTH0_DOMAIN= "login.sandbox.mindgard.ai"
AUTH0_CLIENT_ID= "U0OT7yZLJ4GEyabar11BENeQduu4MaNO"
AUTH0_AUDIENCE="https://marketplace-orchestrator.com"
ALGORITHMS = ['RS256']




import sys
from typing import Any, Dict, Optional, Tuple

import requests

from .constants import REPOSITORY_URL, VERSION


def print_to_stderr(*args: Any, **kwargs: Any) -> None:
    print(*args, file=sys.stderr, **kwargs)


def version_to_tuple(version: str) -> Tuple[int, ...]:
    return tuple(map(int, version.split(".")))


def is_version_outdated() -> Optional[str]:
    try:
        res = requests.get(REPOSITORY_URL)
        res.raise_for_status()
        latest_version = res.json()["info"]["version"]
        latest_version_tuple = version_to_tuple(latest_version)
        current_version_tuple = version_to_tuple(VERSION)
        return latest_version if latest_version_tuple > current_version_tuple else None
    except Exception:
        return ''
    

def api_get(url: str, access_token: str) -> requests.Response:
    res = requests.get(url, headers={
        "Authorization": f"Bearer {access_token}",
        "User-Agent": f"mindgard/{VERSION}"
    })
    res.raise_for_status()
    return res


def api_post(url: str, access_token: str, json: Dict[str, Any]) -> requests.Response:
    res = requests.post(url, headers={
        "Authorization": f"Bearer {access_token}",
        "User-Agent": f"mindgard/{VERSION}"
    }, json=json)
    res.raise_for_status()
    return res






import json
import sys
import time
from typing import Any, Dict, List, Optional

from requests import Response
import requests
from tabulate import tabulate

from .constants import VERSION

from .utils import api_get, api_post

from .auth import require_auth


# TODO: tidy this
def display_test_results(data: List[Dict[str, Any]]) -> None: # TODO: consider color-coded output for risks
    display_data: List[Dict[str, Any]] = []
    for d in data:
        row = {k: v for k, v in d.items() if k not in ["attacks", "updatedAt", "displayName", "source", "isCompleted", "hasFinished"]}
        row["attack_id"] = d["attacks"][0]["id"]
        row["attack_name"] = d["attacks"][0]["attack"]
        row["state"] = d["attacks"][0]["state_message"]
        row["runtime"] = d["attacks"][0]["runtime"] or time.time() - d["attacks"][0]["run_at_unix"]
        row["attack_risk"] = d["attacks"][0]["risk"] if d["attacks"][0]["state_message"] == "Completed" else "TBD"

        display_data.append(row)
        for attack in d["attacks"][1:]:
            display_data.append({
                "attack_id": attack["id"],
                "attack_name": attack["attack"],
                "state": attack["state_message"],
                "runtime": attack["runtime"] or time.time() - attack["run_at_unix"],
                "attack_risk": attack["risk"] if attack["state_message"] == "Completed" else "TBD",
            })
    print(tabulate(display_data, headers="keys"))


@require_auth
def get_tests(access_token: str, json_format: bool = False, test_id: Optional[str] = None) -> Response:
    url = f"https://api.sandbox.mindgard.ai/api/v1/assessments/{test_id}" if test_id else "https://api.sandbox.mindgard.ai/api/v1/assessments?ungrouped=true"
    res = requests.get(url, headers={
        "Authorization": f"Bearer {access_token}",
        "User-Agent": f"mindgard/{VERSION}"
    })
    res.raise_for_status()
    data: List[Dict[str, Any]] = res.json() if isinstance(res.json(), list) else [res.json()]
    if json_format:
        print(json.dumps(res.json()))
    else:
        display_test_results(data)
    return res


@require_auth
def run_test(access_token: str, attack_name: str, json_format: bool = False) -> Response:
    if not json_format:
        print("Initiating testing...")
    res = api_post("https://api.sandbox.mindgard.ai/api/v1/assessments", access_token, {"mindgardModelName": attack_name})
    if json_format:
        print(json.dumps(res.json())) # TODO: include url
    else:
        print("Test initiated. Waiting for results...")
        test_id = res.json()["id"]
        completed = False

        test_res = api_get(f"https://api.sandbox.mindgard.ai/api/v1/assessments/{test_id}", access_token)  
        completed = test_res.json()["hasFinished"]
        display_test_results([test_res.json()])

        num_lines = len(test_res.json()["attacks"]) + 2

        while not completed:
            time.sleep(1)

            sys.stdout.write(f"\033[{num_lines}A") 
            sys.stdout.write("\033[J")

            test_res = api_get(f"https://api.sandbox.mindgard.ai/api/v1/assessments/{test_id}", access_token)  
            completed = test_res.json()["hasFinished"]
            display_test_results([test_res.json()])
        print("Test completed.")
    return res





import os
import time
from typing import Any, Callable, Optional, TypeVar
from functools import wraps

import requests
from auth0.authentication.token_verifier import (AsymmetricSignatureVerifier,
                                                 TokenVerifier)

from .constants import AUTH0_AUDIENCE, AUTH0_CLIENT_ID, AUTH0_DOMAIN
from .utils import print_to_stderr



def get_config_directory() -> str:
    config_dir = os.environ.get('MINDGARD_CONFIG_DIR')
    return config_dir or os.path.join(os.path.expanduser('~'), '.mindgard')


def get_token_file() -> str:
    return os.path.join(get_config_directory(), 'token.txt')


def clear_token() -> None:
    if os.path.exists(get_token_file()):
        os.remove(get_token_file())


# TODO: test
def validate_id_token(id_token: str) -> None:
    """
    Verify the token and its precedence. Raises if the token is invalid.

    :param id_token: string
    """
    jwks_url = 'https://{}/.well-known/jwks.json'.format(AUTH0_DOMAIN)
    issuer = 'https://{}/'.format(AUTH0_DOMAIN)
    sv = AsymmetricSignatureVerifier(jwks_url)
    tv = TokenVerifier(signature_verifier=sv, issuer=issuer, audience=AUTH0_CLIENT_ID)
    tv.verify(id_token)


def load_access_token() -> Optional[str]:
    if os.path.exists(get_token_file()):
        with open(get_token_file(), 'r') as f:
            token = f.read()
            if token:
                return token
    return None
    

def auth() -> None:
    """
    Runs the device authorization flow and stores the user token in memory
    """

    print("Welcome to Mindgard! Let\'s get you authenticated...")

    device_code_payload = {
        'client_id': AUTH0_CLIENT_ID,
        'scope': 'openid profile email',
        'audience': AUTH0_AUDIENCE
    }
    device_code_response = requests.post('https://{}/oauth/device/code'.format(AUTH0_DOMAIN), data=device_code_payload)

    if device_code_response.status_code != 200:
        print_to_stderr('Error generating login url. Please try again. Contact Mindgard support if the issue persists.')
        raise Exception(device_code_response.json())

    device_code_data = device_code_response.json()
    print('1. On your computer or mobile device navigate to: ', device_code_data['verification_uri_complete'])
    print('2. Confirm that you see the following code: ', device_code_data['user_code'])
    print('3. Register/log in using the web UI')


    # New code 👇
    token_payload = {
        'grant_type': 'urn:ietf:params:oauth:grant-type:device_code',
        'device_code': device_code_data['device_code'],
        'client_id': AUTH0_CLIENT_ID,
        'audience': AUTH0_AUDIENCE
    }

    authenticated = False
    while not authenticated:
        print('Checking if the login flow has been completed...    no pressure!')
        token_response = requests.post('https://{}/oauth/token'.format(AUTH0_DOMAIN), data=token_payload)

        token_data = token_response.json()
        if token_response.status_code == 200:
            validate_id_token(token_data['id_token'])
            print('Authenticated!')
            os.makedirs(get_config_directory(), exist_ok=True)
            with open(get_token_file(), 'w') as f:
                f.write(token_data['access_token'])
            authenticated = True
        elif token_data['error'] not in ('authorization_pending', 'slow_down'):
            error = token_data.get('error_description', 'Error authenticating the user. Please wait 30s and try again.')
            raise Exception(error)
        else:
            time.sleep(device_code_data['interval'])
    

T = TypeVar('T')


# TODO: improve typing definitions here
def require_auth(func: Callable[..., T]) -> Callable[..., Optional[T]]:
    @wraps(func)
    def wrapper(*args: Any, **kwargs: Any) -> Optional[T]:
        access_token = load_access_token()
        if not access_token:
            print_to_stderr("First authenticate with Mindgard API.")
            print_to_stderr("Run 'mindgard auth' to authenticate.")
            return None
        try:
            res: T = func(access_token, *args, **kwargs)
        except requests.HTTPError as e:
            if "Unauthorized" in str(e):
                print_to_stderr("Access token is invalid. Please re-authenticate using `mindgard auth`")
                clear_token()
                return None
            else:
                print_to_stderr(f"An error occurred: {e}")
                return None
        except Exception as e:
            print_to_stderr(f"An error occurred: {e}")
            return None
        return res
    return wrapper




import argparse
import sys
from typing import List

from .attacks import attackcategories, get_attacks

from .auth import auth
from .constants import VERSION
from .tests import get_tests, run_test
from .utils import is_version_outdated, print_to_stderr


def parse_args(args: List[str]) -> argparse.Namespace:
    parser = argparse.ArgumentParser(description='Securing AIs', prog='mindgard', usage='%(prog)s [command] [options]', epilog='Enjoy the program! :)', add_help=True)
    parser.add_argument('--version', action='version', version=f"%(prog)s {VERSION}", help='Show the current version number')

    subparsers = parser.add_subparsers(dest='command', title='commands', description='Use these commands to interact with the Mindgard API')

    attack_categories_parser = subparsers.add_parser('attackcategories', help='Get a list of attack categories.')
    attack_categories_parser.add_argument('--json', action="store_true", help='Output the info in JSON format.') 

    subparsers.add_parser('auth', help='Authenticate with Mindgard API')

    # TODO: think about more streamlined command for running a test
    test_parser = subparsers.add_parser('tests', help='See the tests you\'ve run.') # TODO: better help text
    test_parser.add_argument('--json', action="store_true", help='Output the info in JSON format.', required=False) # TODO: don't allow this if run comes after
    test_parser.add_argument('--id', type=str, help='Get the details of a specific test.', required=False)
    
    test_subparsers = test_parser.add_subparsers(dest='test_commands', title='test_commands', description='Perform actions against tests')
    test_run_parser = test_subparsers.add_parser('run', help='Run a test.') # TODO: risk exit codes
    # TODO: links to view results in the UI for images etc
    test_run_parser.add_argument('--name', type=str, help='The attack to test.', required=True, choices=['cfp_faces', 'mistral'])
    test_run_parser.add_argument('--json', action="store_true", help='Initiate test and return id that can be used to check status.', required=False)

    attack_parser = subparsers.add_parser('attacks', help='See the attacks you\'ve run.')
    attack_parser.add_argument('--json', action="store_true", help='Output the info in JSON format.', required=False)
    attack_parser.add_argument('--id', type=str, help='Get the details of a specific attack.', required=False)

    return parser.parse_args(args)
    


def main() -> None:
    args = parse_args(sys.argv[1:])

    if not (sys.version_info.major == 3 and sys.version_info.minor >= 8):
        print_to_stderr("Python 3.8 or later is required to run the Mindgard CLI.")
        sys.exit(1)

    if new_version := is_version_outdated():
        print_to_stderr(f"New version available: {new_version}. Run 'pip install mindgard --upgrade' to upgrade. Older versions of the CLI may not be actively maintained.")

    if args.command == 'auth':
        auth()
    elif args.command == 'attackcategories':
        attackcategories(json_format=args.json)
    elif args.command == 'tests':
        if args.test_commands == "run":
            run_test(attack_name=args.name, json_format=bool(args.json))
        else:
            get_tests(json_format=bool(args.json), test_id=args.id)
    elif args.command == 'attacks':
        get_attacks(json_format=args.json, attack_id=args.id)
    else:
        print_to_stderr('Hey give us a command. Use list or auth.') # TODO update


if __name__ == '__main__':
    main()